OP_CATEGORY exceptions, "Exception Handling (FEATURE__EXCEPTIONS)", "These are non-local returns that can also be used for error handling."
OP_CATDOC exceptions, {"When a throw is triggered, the CPU and register stacks are restored to the state at the time of the catch, the register stack grows by 2 slots, and r0 & r1 become the exception tag and parameter, respectively, with rP pointing to r0.", "  Usually a handler will use 'case' instructions to branch on desired tags, rethrowing the exception otherwise."}
OP_CATDOC exceptions, "For handling 'finally' situations, normal code flow and exception rethrowing need to be handled in the same code:"
OP_CATDOC exceptions, {"<pre>  catch finally",EOL,"  ...",EOL,"  popcatch",EOL,"  grow 2",EOL,"  set r0, 0",EOL,"finally:",EOL,"  ...",EOL,"  throw r0,r1  ; ignored if no exception was thrown and r0 is still zero</pre>"}
OP_CATDOC exceptions, "Each catch context takes 5 bytes on the CPU stack."

ZPVAR currentCatch, exceptions, 1, "CPU stack position describing the currently registered exception handler."

OP catch, imm16, exceptions, "Register an exception handler routine at imm16."
 lda currentCatch
 pha
 lda (iptr),y ; low
 iny
 pha
 lda (iptr),y ; high
 pha
 lda rptr
 pha
 tsx
 stx currentCatch
 jmp mainLoop3

OP popcatch, none, exceptions, "Discard the most recent exception handler."
 tsx
 txa
 adc #4
 tax
 txs
 pla
 sta currentCatch
 jmp mainLoop1

:jmp mainLoop2
OP throw, rdra, exceptions, "If rD is nonzero, throw an exception with tag rD and parameter rA.  Can be used to rethrow from inside a catch handler."
_exceptionTrapVector = *+1
 jsr _decode_rdra   ; selfmodded by monitor hooks
_exceptionTrapResume:
 lda 0,x
 ora 1,x
 beq :-
 stx zptemp
 ldx currentCatch
 txs
 pla
 sec
 sbc #4        ; grow by 2 slots.  TODO - overflow check
 sta rptr
 sta pptr
 pla
 sta iptr+1
 pla
 sta iptr
 pla
 sta currentCatch
 jsr copyZYTor0r1
 jmp mainLoop0
