;-------------------
; Trap support

OP_CATEGORY traps, "Monitor Traps (FEATURE__TRAPS)", "The trap system allows breaking into native code before each instruction is executed, or when exceptions are thrown.  Native code can then check the iptr for breakpoint matches, resume, single-step, swap tasks, or do whatever it wants."

OP_CATDOC traps, "When the trap runs, the iptr points to the beginning of the instruction yet to be executed."
OP_CATDOC traps, {"These handlers self-modify the main loop, so there is no runtime overhead when this feature is included and disabled, besides the memory footprint.","  Since the selfmod happens only on native instruction boundaries, it is safe to enable/disable traps from interrupt handlers.","  Preemptive task switchers should use this sort of approach."}

NATIVE enableInstructionTrap, traps, jsr, "Enable the instruction trap to call <.A >.X."
 sta _instructionTrapJump +1
 stx _instructionTrapJump +2
 lda #$f0 ;beq
 sta _instructionTrapSelfmod
 lda #_instructionTrapJump - _instructionTrapSelfmod - 2
 sta _instructionTrapSelfmod + 1
 rts

NATIVE disableInstructionTrap, traps, jsr, "Disable the break functionality."
 lda #$b1 ;lda (zp),y
 sta _instructionTrapSelfmod
 lda #iptr
 sta _instructionTrapSelfmod+1
 rts

NATIVE continueFromInstructionTrap, traps, jmp, "Continue running the VM after a break was triggered.  If the break is still enabled, this effectively single-steps."
 ; Simulate what the selfmod stuff did
 ldy #0
 lda (iptr),y
 jmp _instructionTrapResume


.if FEATURE__EXCEPTIONS
NATIVE enableExceptionTrap, traps, jsr, "Enable the exception trap to call <.A >.X"
 sta _exceptionTrapVector
 stx _exceptionTrapVector+1
 rts

NATIVE disableExceptionTrap, traps, jsr, "Disable exception trap."
 lda #<_decode_rdra
 sta _exceptionTrapVector
 lda #>_decode_rdra
 sta _exceptionTrapVector+1
 rts

NATIVE continueFromExceptionTrap, traps, jmp, "Continue processing the exception thrown."
 ldy #0
 jsr _decode_rdra
 jmp _exceptionTrapResume

.endif