;-------------------------
; Output helpers

.define HEXCHAR(b) (('0'+(b))*((b)<=9) + ('a'+(b)-10)*((b)>9))
.define HEXBYTE(b) HEXCHAR(((b)>>4)&$f), HEXCHAR((b)&$f)
.define HEXWORD(b) HEXCHAR(((b)>>12)&$f), HEXCHAR(((b)>>8)&$f), HEXCHAR(((b)>>4)&$f), HEXCHAR((b)&$f)

.macro indent var, args, size
  .repeat size-.strlen(args)-.strlen(.string(var))
   .byte " "
  .endrepeat
.endmacro



;-------------------------
; File headers

.segment "INCFILE_TOP"
 .byte "; Autogenerated AcheronVM instruction definitions",EOL, "; See acheron.asm for source",EOL,EOL
.macro defreg n
 .byte ".define ", .sprintf("r%d %d",n,n),EOL
.endmacro
.repeat 16, n
 defreg n
.endrepeat
 .byte EOL

.segment "OPDESC_"
 .byte "(",EOL

.if (.xmatch(PLATFORM_ENCODING, cbm_prg))
 .segment "BIN_HEADER"
 .word unembeddedDispatchTable
.endif


;-------------------------
; Opcode generation

.if COMPACT_UNEMBEDDED_OPCODES
 .define opcodesPerUnembeddedInstruction 1
 .define convertUnembeddedToDispatch asl
.else
 .define opcodesPerUnembeddedInstruction 2
 .define convertUnembeddedToDispatch
.endif

.segment "UTABLE"
unembeddedDispatchTable:
.segment "ETABLE"
embeddedDispatchTable:

.macro outputOpcode name, embed, prefix
 .ifblank prefix
  .byte '$'
 .else
  .byte prefix
 .endif
 .if (.xmatch(embed,u))
  .byte HEXBYTE((_entry - unembeddedDispatchTable)/2 * opcodesPerUnembeddedInstruction)
 .elseif (.xmatch(embed,e))
  .byte HEXBYTE(((__firstEmbeddedOpcode >> 4) + ((_entry - embeddedDispatchTable) >> 1)) << 4)
 .else
  .error .concat("Instruction ",.string(name)," should use e or u for its type, not ",.string(embed))
 .endif
.endmacro

.macro EMIT embed, name, category, doc, params, bytespec, paramspecdoc
 .pushseg
 .segment "INCFILE_OPCODES"
 .scope
  .byte ".define opcode_", .string(name), " "
  indent name, "", 20
  outputOpcode name, embed
  .byte EOL
  .segment "INCFILE_INSTRUCTIONS"
  .byte ".define ", .string(name), " (", params, ")"
  indent name, params, 26
  .byte ".byte "
  outputOpcode name, embed
 .endscope
 .byte bytespec,EOL
 .segment "OPDESC_"
 .byte " (op (embed ",.string(embed),") (opcode "
 outputOpcode name, embed, "#x"
 .byte ") (name ",'"',.string(name),'"',") (category ",.string(category),") (paramtext ",'"',params,'"',") (doc ",'"',doc,'"',"))",EOL
 .popseg
.endmacro




;-------------------------
; Zeropage management

.import __ZPSTACK_START__, __ZPSTACK_SIZE__

.define CURRENT_ZP_LOC <(zpTop + 1 + __ZPSTACK_START__ - *)   ; * is start + delta, we want start + size - delta

.macro ZPVAR name, category, size, doc, nolabel
 .exportzp name
 .pushseg
 .ifnblank nolabel
 .else
  .segment "ZPSTACK": zeropage
  .res size
  name = CURRENT_ZP_LOC
 .endif
 .segment "INCFILE_IMPORTS"
 .byte ".importzp ",.string(name),EOL
 .segment "OPDESC_"
 .byte " (zeropage (name ",'"',.string(name),'"',") (category ",.string(category),") (size ",.string(size),") (doc ",'"',doc,'"',"))",EOL
 .popseg
.endmacro

ZPVAR zpTop, regs, 0, "Highest zeropage memory location in use, for copying out process context.", nolabel
zpTop = <(__ZPSTACK_START__ + __ZPSTACK_SIZE__ - 1)




;-------------------------
; Declarations

.macro PSEUDO name, params, category, doc, contents
 .pushseg
 .segment "OPDESC_"
 .byte " (pseudo-op (name ",'"',.string(name),'"',") (category ",.string(category),") (params ",'"',params,'"',") (doc ",'"',doc,'"',"))",EOL
 .segment "INCFILE_PSEUDO"
 .byte ".macro ",.string(name), " ", params,EOL, contents,EOL, ".endmacro",EOL,EOL
 .popseg
.endmacro

.macro NATIVE name, category, calltype, doc, nolabel
 .ifnblank nolabel
 .else
 .export name
 name:
 .endif
 .pushseg
 .segment "OPDESC_"
 .byte " (native (name ",'"',.string(name),'"',") (calltype ",'"',.string(calltype),'"',") (category ",.string(category),") (doc ",'"',doc,'"',"))",EOL
 .segment "INCFILE_IMPORTS"
 .byte ".import ",.string(name),EOL
 .popseg
.endmacro

.macro OP_CATEGORY name, title, header
 .pushseg
 .segment "OPDESC_"
 .byte " (category (name ",.string(name),") (title ",'"',title,'"',")"
 .ifnblank header
   .byte " (header ",'"',header,'"',")"
 .endif
 .byte ")",EOL
 .popseg
.endmacro

; Additional documentation at the end of the category.
; This can be called multiple times, each call creates a new paragraph.
; If the doc paragraph is broken up between multiple strings, enclose it all with curly braces.
.macro OP_CATDOC name, doc
 .pushseg
 .segment "OPDESC_"
 .byte " (category-doc (category ",.string(name),") (doc ",'"',doc,'"',"))",EOL
 .popseg
.endmacro




;-------------------------
; Parameter decoding

.macro decode_rdra
 jsr _decode_rdra
.endmacro

.macro decode_rdimm4
 jsr _decode_rdimm4
.endmacro

.macro decode_rd
 lda (iptr),y
 adc rptr
 sta pptr
 tax
.endmacro
